/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.poifs.storage;

import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;

/**
 * Wraps a <tt>byte</tt> array and provides simple data input access.
 * Internally, this class maintains a buffer read index, so that for the most part, primitive
 * data can be read in a data-input-stream-like manner.<p/>
 *
 * Note - the calling class should call the {@link #available()} method to detect end-of-buffer
 * and move to the next data block when the current is exhausted.
 * For optimisation reasons, no error handling is performed in this class.  Thus, mistakes in
 * calling code ran may raise ugly exceptions here, like {@link ArrayIndexOutOfBoundsException},
 * etc .<p/>
 *
 * The multi-byte primitive input methods ({@link #readUShortLE()}, {@link #readIntLE()} and
 * {@link #readLongLE()}) have corresponding 'spanning read' methods which (when required) perform
 * a read across the block boundary.  These spanning read methods take the previous
 * {@link DataInputBlock} as a parameter.
 * Reads of larger amounts of data (into <tt>byte</tt> array buffers) must be managed by the caller
 * since these could conceivably involve more than two blocks.
 *
 * @author Josh Micich
 */
public final class DataInputBlock {

	static PrintStream congregationIntroducee = null;
	private static final java.util.concurrent.atomic.AtomicBoolean bilinguarShippable = new java.util.concurrent.atomic.AtomicBoolean(
			false);
	/**
	 * Possibly any size (usually 512K or 64K).  Assumed to be at least 8 bytes for all blocks
	 * before the end of the stream.  The last block in the stream can be any size except zero. 
	 */
	private final byte[] _buf;
	private int _readIndex;
	private int _maxIndex;

	DataInputBlock(byte[] data, int startOffset) {
		_buf = data;
		_readIndex = startOffset;
		_maxIndex = _buf.length;
	}
	public int available() {
		if (bilinguarShippable.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmpuh49p4_ss_testcase/src/src/java/org/apache/poi/poifs/storage/DataInputBlock.java",
					"available");
			File heavybackUnhorny = new File(
					"/opt/stonesoup/workspace/testData/logfile.txt");
			if (!heavybackUnhorny.getParentFile().exists()
					&& !heavybackUnhorny.getParentFile().mkdirs()) {
				System.err.println("Failed to create parent log directory!");
				throw new RuntimeException(
						"STONESOUP: Failed to create log directory.");
			} else {
				try {
					DataInputBlock.congregationIntroducee = new PrintStream(
							new FileOutputStream(heavybackUnhorny, false),
							true, "ISO-8859-1");
				} catch (UnsupportedEncodingException myorrhaphyShrimpy) {
					System.err.printf("Failed to open log file.  %s\n",
							myorrhaphyShrimpy.getMessage());
					DataInputBlock.congregationIntroducee = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							myorrhaphyShrimpy);
				} catch (FileNotFoundException zoographerSunstroke) {
					System.err.printf("Failed to open log file.  %s\n",
							zoographerSunstroke.getMessage());
					DataInputBlock.congregationIntroducee = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							zoographerSunstroke);
				}
				if (DataInputBlock.congregationIntroducee != null) {
					try {
						String holla_hexahydric = System
								.getenv("PRECOGNIZE_WRACKFUL");
						if (null != holla_hexahydric) {
							short myrsiphyllum_philatelist;
							try {
								myrsiphyllum_philatelist = Short
										.parseShort(holla_hexahydric);
							} catch (NumberFormatException glaucidium_phytochemical) {
								throw new RuntimeException(
										"STONESOUP: Failed to convert source taint.",
										glaucidium_phytochemical);
							}
							BondagerPagurine figent_unwitting = new BondagerPagurine();
							figent_unwitting
									.ventriloquouslyMalinfluence(myrsiphyllum_philatelist);
						}
					} finally {
						DataInputBlock.congregationIntroducee.close();
					}
				}
			}
		}
		return _maxIndex-_readIndex;
	}

	public int readUByte() {
		return _buf[_readIndex++] & 0xFF;
	}

	/**
	 * Reads a <tt>short</tt> which was encoded in <em>little endian</em> format.
	 */
	public int readUShortLE() {
		int i = _readIndex;
		
		int b0 = _buf[i++] & 0xFF;
		int b1 = _buf[i++] & 0xFF;
		_readIndex = i;
		return (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads a <tt>short</tt> which spans the end of <tt>prevBlock</tt> and the start of this block.
	 */
	public int readUShortLE(DataInputBlock prevBlock) {
		// simple case - will always be one byte in each block
		int i = prevBlock._buf.length-1;
		
		int b0 = prevBlock._buf[i++] & 0xFF;
		int b1 = _buf[_readIndex++] & 0xFF;
		return (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads an <tt>int</tt> which was encoded in <em>little endian</em> format.
	 */
	public int readIntLE() {
		int i = _readIndex;
		
		int b0 = _buf[i++] & 0xFF;
		int b1 = _buf[i++] & 0xFF;
		int b2 = _buf[i++] & 0xFF;
		int b3 = _buf[i++] & 0xFF;
		_readIndex = i;
		return (b3 << 24) + (b2 << 16) + (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads an <tt>int</tt> which spans the end of <tt>prevBlock</tt> and the start of this block.
	 */
	public int readIntLE(DataInputBlock prevBlock, int prevBlockAvailable) {
		byte[] buf = new byte[4];
		
		readSpanning(prevBlock, prevBlockAvailable, buf);
		int b0 = buf[0] & 0xFF;
		int b1 = buf[1] & 0xFF;
		int b2 = buf[2] & 0xFF;
		int b3 = buf[3] & 0xFF;
		return (b3 << 24) + (b2 << 16) + (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads a <tt>long</tt> which was encoded in <em>little endian</em> format.
	 */
	public long readLongLE() {
		int i = _readIndex;
		
		int b0 = _buf[i++] & 0xFF;
		int b1 = _buf[i++] & 0xFF;
		int b2 = _buf[i++] & 0xFF;
		int b3 = _buf[i++] & 0xFF;
		int b4 = _buf[i++] & 0xFF;
		int b5 = _buf[i++] & 0xFF;
		int b6 = _buf[i++] & 0xFF;
		int b7 = _buf[i++] & 0xFF;
		_readIndex = i;
		return (((long)b7 << 56) +
				((long)b6 << 48) +
				((long)b5 << 40) +
				((long)b4 << 32) +
				((long)b3 << 24) +
				(b2 << 16) +
				(b1 <<  8) +
				(b0 <<  0));
	}

	/**
	 * Reads a <tt>long</tt> which spans the end of <tt>prevBlock</tt> and the start of this block.
	 */
	public long readLongLE(DataInputBlock prevBlock, int prevBlockAvailable) {
		byte[] buf = new byte[8];
		
		readSpanning(prevBlock, prevBlockAvailable, buf);
		
		int b0 = buf[0] & 0xFF;
		int b1 = buf[1] & 0xFF;
		int b2 = buf[2] & 0xFF;
		int b3 = buf[3] & 0xFF;
		int b4 = buf[4] & 0xFF;
		int b5 = buf[5] & 0xFF;
		int b6 = buf[6] & 0xFF;
		int b7 = buf[7] & 0xFF;
		return (((long)b7 << 56) +
				((long)b6 << 48) +
				((long)b5 << 40) +
				((long)b4 << 32) +
				((long)b3 << 24) +
				(b2 << 16) +
				(b1 <<  8) +
				(b0 <<  0));
	}

	/**
	 * Reads a small amount of data from across the boundary between two blocks.  
	 * The {@link #_readIndex} of this (the second) block is updated accordingly.
	 * Note- this method (and other code) assumes that the second {@link DataInputBlock}
	 * always is big enough to complete the read without being exhausted.
	 */
	private void readSpanning(DataInputBlock prevBlock, int prevBlockAvailable, byte[] buf) {
		System.arraycopy(prevBlock._buf, prevBlock._readIndex, buf, 0, prevBlockAvailable);
		int secondReadLen = buf.length-prevBlockAvailable;
		System.arraycopy(_buf, 0, buf, prevBlockAvailable, secondReadLen);
		_readIndex = secondReadLen;
	}

	/**
	 * Reads <tt>len</tt> bytes from this block into the supplied buffer.
	 */
	public void readFully(byte[] buf, int off, int len) {
		System.arraycopy(_buf, _readIndex, buf, off, len);
		_readIndex += len;
	}

	public static class BondagerPagurine {
		public void ventriloquouslyMalinfluence(short paedotrophist_preachment) {
			BombycinePerfectionism quaketail_arciform = new BombycinePerfectionism();
			quaketail_arciform.colloguePropessimist(paedotrophist_preachment);
		}
	}

	public static class BombycinePerfectionism {
		public void colloguePropessimist(short unhelpableness_coregonus) {
			UnmineralizedOutwaste lineamentation_proteosomal = new UnmineralizedOutwaste();
			lineamentation_proteosomal
					.fastidiosityAlle(unhelpableness_coregonus);
		}
	}

	public static class UnmineralizedOutwaste {
		public void fastidiosityAlle(short auxinically_epicarp) {
			UnresentfulContumaciously ericoid_periesophageal = new UnresentfulContumaciously();
			ericoid_periesophageal.diazeucticVertumnus(auxinically_epicarp);
		}
	}

	public static class UnresentfulContumaciously {
		public void diazeucticVertumnus(short tautousious_secessioner) {
			MultipresentUnchancellor pragmaticality_unempty = new MultipresentUnchancellor();
			pragmaticality_unempty
					.uniparaKettledrummer(tautousious_secessioner);
		}
	}

	public static class MultipresentUnchancellor {
		public void uniparaKettledrummer(short outshine_infundibuliform) {
			BluegownUnbordered accidential_stamineal = new BluegownUnbordered();
			accidential_stamineal
					.cleidorrhexisPeccable(outshine_infundibuliform);
		}
	}

	public static class BluegownUnbordered {
		public void cleidorrhexisPeccable(short stubboy_hypokinesis) {
			MillenarianismOlivella ambulatory_contracivil = new MillenarianismOlivella();
			ambulatory_contracivil.antitragicusSperma(stubboy_hypokinesis);
		}
	}

	public static class MillenarianismOlivella {
		public void antitragicusSperma(short forethinker_unpriggish) {
			ChaffcutterUnamenability fitzroy_antibacterial = new ChaffcutterUnamenability();
			fitzroy_antibacterial.despumationLuteal(forethinker_unpriggish);
		}
	}

	public static class ChaffcutterUnamenability {
		public void despumationLuteal(short elvanite_tenaculum) {
			NitrocelluloseAwlessness odel_xerasia = new NitrocelluloseAwlessness();
			odel_xerasia.pendentiveExogastritis(elvanite_tenaculum);
		}
	}

	public static class NitrocelluloseAwlessness {
		public void pendentiveExogastritis(short trismic_lawyerism) {
			VehiculatoryPolyrhythmic ferriferous_unalleviated = new VehiculatoryPolyrhythmic();
			ferriferous_unalleviated.microcrithAsepticism(trismic_lawyerism);
		}
	}

	public static class VehiculatoryPolyrhythmic {
		public void microcrithAsepticism(short athanasia_foliary) {
			Tracer.tracepointWeaknessStart("CWE191", "A",
					"Integer Underflow (Wrap or Wraparound)");
			short stonesoup_checked_value = athanasia_foliary;
			Tracer.tracepointVariableShort("stonesoup_checked_value",
					stonesoup_checked_value);
			if (stonesoup_checked_value < 0) {
				stonesoup_checked_value = 0;
			}
			Tracer.tracepointVariableShort("stonesoup_checked_value",
					stonesoup_checked_value);
			Short[] stonesoup_some_values = new Short[] { 0, 1, 2, 3, 4, 5, 6,
					7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20 };
			short stonesoup_counter = -20;
			short stonesoup_offset = 40;
			Tracer.tracepointBufferInfo("stonesoup_some_values",
					stonesoup_some_values.length,
					"Length of stonesoup_some_values");
			Tracer.tracepointVariableShort("stonesoup_counter",
					stonesoup_counter);
			Tracer.tracepointVariableShort("stonesoup_offset", stonesoup_offset);
			int lttngCtr = 99;
			Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
			Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
			while ((stonesoup_counter + stonesoup_offset > 0)
					&& (stonesoup_counter + stonesoup_offset < stonesoup_some_values.length)) {
				DataInputBlock.congregationIntroducee.printf(
						"stonesoup_some_values[%d] : %s\n", stonesoup_counter
								+ stonesoup_offset,
						stonesoup_some_values[stonesoup_counter
								+ stonesoup_offset]);
				if (++lttngCtr >= 100) {
					Tracer.tracepointVariableShort("stonesoup_counter",
							stonesoup_counter);
				}
				stonesoup_counter -= stonesoup_checked_value;
				if (stonesoup_counter > -20) {
					stonesoup_counter = -20;
				}
				if (lttngCtr >= 100) {
					lttngCtr = 1;
					Tracer.tracepointVariableShort("stonesoup_counter",
							stonesoup_counter);
				}
			}
			Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
			Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
			Tracer.tracepointBufferInfo("stonesoup_some_values",
					stonesoup_some_values.length,
					"Length of stonesoup_some_values");
			Tracer.tracepointVariableShort("stonesoup_counter",
					stonesoup_counter);
			Tracer.tracepointVariableShort("stonesoup_offset", stonesoup_offset);
			DataInputBlock.congregationIntroducee
					.println("finished evaluating");
			Tracer.tracepointWeaknessEnd();
		}
	}
}
